from functools import wraps


class weight(object):
    """Simple decorator to add a __weight__ property to a function

    Usage: @weight(3.0)
    """
    def __init__(self, val):
        self.val = val

    def __call__(self, func):
        func.__weight__ = self.val
        return func


class number(object):
    """Simple decorator to add a __number__ property to a function

    Usage: @number("1.1")

    This field will then be used to sort the test results on Gradescope.
    """

    def __init__(self, val):
        self.val = val

    def __call__(self, func):
        func.__number__ = self.val
        return func

class visibility(object):
    """Simple decorator to add a __visibility__ property to a function

    Usage: @visibility("hidden")

    Options for the visibility field are as follows:

    - `hidden`: test case will never be shown to students
    - `after_due_date`: test case will be shown after the assignment's due date has passed.
      If late submission is allowed, then test will be shown only after the late due date.
    - `after_published`: test case will be shown only when the assignment is explicitly published from the "Review Grades" page
    - `visible` (default): test case will always be shown
    """

    def __init__(self, val):
        self.val = val

    def __call__(self, func):
        func.__visibility__ = self.val
        return func


class tags(object):
    """Simple decorator to add a __tags__ property to a function

    Usage: @tags("concept1", "concept2")
    """
    def __init__(self, *args):
        self.tags = args

    def __call__(self, func):
        func.__tags__ = self.tags
        return func


class leaderboard(object):
    """Decorator that indicates that a test corresponds to a leaderboard column

    Usage: @leaderboard("high_score"). The string parameter indicates
    the name of the column on the leaderboard

    Then, within the test, set the value by calling
    kwargs['set_leaderboard_value'] with a value. You can make this convenient by
    explicitly declaring a set_leaderboard_value keyword argument, eg.

    ```
    def test_highscore(set_leaderboard_value=None):
        set_leaderboard_value(42)
    ```

    """

    def __init__(self, column_name, sort_order='desc'):
        self.column_name = column_name
        self.sort_order = sort_order

    def __call__(self, func):
        func.__leaderboard_column__ = self.column_name
        func.__leaderboard_sort_order__ = self.sort_order

        def set_leaderboard_value(x):
            wrapper.__leaderboard_value__ = x

        @wraps(func)
        def wrapper(*args, **kwargs):
            kwargs['set_leaderboard_value'] = set_leaderboard_value
            return func(*args, **kwargs)

        return wrapper


class partial_credit(object):
    """Decorator that indicates that a test allows partial credit

    Usage: @partial_credit(test_weight)

    Then, within the test, set the value by calling
    kwargs['set_score'] with a value. You can make this convenient by
    explicitly declaring a set_score keyword argument, eg.

    ```
    @partial_credit(10)
    def test_partial(set_score=None):
        set_score(4.2)
    ```

    """

    def __init__(self, weight):
        self.weight = weight

    def __call__(self, func):
        func.__weight__ = self.weight

        def set_score(x):
            wrapper.__score__ = x

        @wraps(func)
        def wrapper(*args, **kwargs):
            kwargs['set_score'] = set_score
            return func(*args, **kwargs)

        return wrapper



import datetime
import time
from typing import List

from . import Autograder
from .Utils import is_local

class RateLimit:
    def __init__(
        self,
        tokens: int=None,
        seconds: int=0,
        minutes: int=0,
        hours: int=0,
        days: int=0,
        reset_time: str=None,
        pull_prev_run: bool=False,
        submission_id_exclude: List[str]=[],
        verbose: bool=False
    ):
        self.tokens = tokens
        self.seconds = seconds
        self.minutes = minutes
        self.hours = hours
        self.days = days
        self.reset_time = reset_time
        self.output = ""

        self.pull_prev_run = pull_prev_run

        self.submission_id_exclude = submission_id_exclude

        self.verbose = verbose

        self.oldest_token_time = None
        self.current_submission_time = None

        self.main_string = ""
        self.tokens_used = ""

    def is_enabled(self):
        return not (self.tokens is None or self.tokens <= 0)

    def print(self, *args, sep=' ', end='\n', file=None, flush=True, also_stdout=False):
        msg = sep.join(map(str, args)) + end
        if also_stdout:
            print(msg)
        self.output += msg

    def set_next_token_regen(self, oldest_token_time, current_submission_time):
        self.oldest_token_time = oldest_token_time
        self.current_submission_time = current_submission_time

    def rate_limit_set_main_string(self, string, tokens_used):
        self.main_string = string
        self.tokens_used = tokens_used

    def get_rate_limit_str(self, ag: Autograder):
        if not self.is_enabled():
            return ""
        datetime_regen_rate = datetime.timedelta(seconds=self.total_seconds())
        sub_to_count = None
        if self.oldest_token_time:
            sub_to_count = self.oldest_token_time
        else:
            if self.rate_limit_does_submission_count(ag):
                sub_to_count = self.current_submission_time
        

        tu = self.tokens_used
        if not self.rate_limit_does_submission_count(ag):
            tu -= 1

        if sub_to_count is not None:
            next_token_regen = sub_to_count + datetime_regen_rate
            next_token_regen_str = f"[Rate Limit]: As of this submission time, your next token will regenerate at {next_token_regen.ctime()} (PT).\n\n"
        else:
            next_token_regen_str = "[Rate Limit]: As of this submission time, you have not used any tokens!\n\n"

        return self.main_string.format(tu) + next_token_regen_str

    def total_seconds(self):
        return self.seconds + 60 * (self.minutes + 60 * (self.hours + (24 * self.days)))

    def rate_limit_main(self, ag: Autograder, verbose=None):
        if verbose is None:
            verbose = self.verbose
        if not self.is_enabled():
            return
        if is_local() and not ag.use_ratelimit_when_local:
            print("[WARNING]: Rate limit is enabled but will not be checked because this has been detected to be a local run!")
            return
        tokens = self.tokens
        restart_subm_string = self.reset_time
        s = self.seconds
        m = self.minutes
        h = self.hours
        d = self.days
        regen_time_seconds = self.total_seconds()
        def get_submission_time(s):
            return s[:-13]
        def pretty_time_str(s, m, h, d):
            sstr = "" if s == 0 else str(s) + " second"
            sstr += "" if sstr == "" or s == 1 else "s"
            mstr = "" if m == 0 else str(m) + " minute"
            mstr += "" if mstr == "" or m == 1 else "s"
            hstr = "" if h == 0 else str(h) + " hour"
            hstr += "" if hstr == "" or h == 1 else "s"
            dstr = "" if d == 0 else str(d) + " day"
            dstr += "" if dstr == "" or d == 1 else "s"
            st = dstr
            for tmpstr in [hstr, mstr, sstr]:
                if st != "" and tmpstr != "":
                    st += " "
                st += tmpstr
            if st == "":
                st = "none"
            return st
        current_subm_string = get_submission_time(ag.metadata["created_at"])
        current_time = time.strptime(current_subm_string,"%Y-%m-%dT%H:%M:%S")
        restart_time = time.strptime(restart_subm_string, "%Y-%m-%dT%H:%M:%S") if restart_subm_string is not None else None
        tokens_used = 0
        if verbose:
            print("=" * 30)
        oldest_counted_submission = None
        for i, v in enumerate(ag.metadata["previous_submissions"]):
            subm_string = get_submission_time(v["submission_time"])
            subm_time = time.strptime(subm_string,"%Y-%m-%dT%H:%M:%S")
            if restart_time is not None and time.mktime(subm_time) - time.mktime(restart_time) < 0:
                if verbose:
                    print("Ignoring a submission, too early!")
                continue
            if verbose:
                print("Current time: " + str(time.mktime(current_time)))
                print("Subm time: " + str(time.mktime(subm_time)))
            if (time.mktime(current_time) - time.mktime(subm_time) < regen_time_seconds): 
                try:
                    if verbose:
                        print(ag.metadata["previous_submissions"][i])
                        print("Tokens used: " + str(tokens_used))
                        print(str(ag.metadata["previous_submissions"][i].keys()))
                        print("Current submission data: " + str(ag.metadata["previous_submissions"][i]["results"]["extra_data"]))
                    ed = ag.metadata["previous_submissions"][i]["results"]["extra_data"]
                    if ed is not None:
                        subID = ed.get("id")
                        if (ed["sub_counts"] == 1) and (subID and (subID not in self.submission_id_exclude)): 
                            if oldest_counted_submission is None:
                                oldest_counted_submission = subm_time
                            tokens_used = tokens_used + 1
                    else:
                        if verbose:
                            print(f"Extra data not available in previous submission {i}!")
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    print(e)
                    tokens_used = tokens_used + 1
                    pass
            if verbose:
                print("-" * 30)
        if verbose:
            print("=" * 30)
        datetime_regen_rate = datetime.timedelta(seconds=regen_time_seconds)
        if oldest_counted_submission:
            oldest_counted_submission = datetime.datetime.fromtimestamp(time.mktime(oldest_counted_submission))
        if current_time:
            datetime_current_time = datetime.datetime.fromtimestamp(time.mktime(current_time))
        if tokens_used < tokens:
            ag.extra_data["sub_counts"] = 1
            tokens_used += 1 # This is to include the current submission.
            self.rate_limit_set_main_string(f"[Rate Limit]: Students can get up to {tokens} graded submissions within any given period of {pretty_time_str(s, m, h, d)}. In the last period, you have had {{}} graded submissions.\n", tokens_used)
            self.set_next_token_regen(oldest_counted_submission, datetime_current_time)
        else:
            ag.extra_data["sub_counts"] = 0
            if self.pull_prev_run:
                msg = ", so the results of your last graded submission are being displayed."
            else:
                msg = "."
            ag.print(f"[Rate Limit]: Students can get up to {tokens} graded submissions within any given period of {pretty_time_str(s, m, h, d)}. You have already had {tokens_used} graded submissions within the last {pretty_time_str(s, m, h, d)}{msg} Because you do not have any more tokens, this submission will not count as a graded submission.")

            if oldest_counted_submission:
                next_token_regen = oldest_counted_submission + datetime_regen_rate
                ag.print(f"[Rate Limit]: As of this submission time, your next token will regenerate at {next_token_regen.ctime()} (PT).\n")
            else:
                ag.print(f"[Rate Limit]: As of this submisison, you have not used any tokens.\n")
            
            if self.pull_prev_run:
                prev_subs = ag.metadata["previous_submissions"]
                prev_sub = prev_subs[len(prev_subs) - 1]
                if prev_sub and "results" not in prev_sub or prev_sub["results"] and "tests" not in prev_sub["results"]:
                    ag.print("[ERROR]: Could not pull the data from your previous submission! This is probably due to it not have finished running!")
                    tests = []
                    ag.set_score(0)
                    leaderboard = None
                else:
                    res = prev_sub["results"]
                    tests = res["tests"]
                    leaderboard = res["leaderboard"]
                    ag.set_score(prev_sub.get("score"))
            else:
                    tests = []
                    ag.set_score(0)
                    leaderboard = None
                
            ag.generate_results(test_results=tests, leaderboard=leaderboard)
            
            import sys
            sys.exit()
            # raise AutograderHalt("Rate limited!")

    @staticmethod
    def rate_limit_unset_submission(ag: Autograder):
        ag.extra_data["sub_counts"] = 0

    @staticmethod
    def rate_limit_does_submission_count(ag: Autograder):
        return ag.extra_data["sub_counts"]


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 20:56:28 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-30 16:13:26
 */
 """
from .Autograder import Autograder, RateLimit
from .AutograderTest import AutograderTest, Max, global_tests
from .AutograderSetup import AutograderSetup
from .AutograderTeardown import AutograderTeardown
from .AutograderSubTest import AutograderSubTest, SubTestRunner, StopSubTestRunner
from .AutograderErrors import AutograderBaseError, AutograderFormatError, AutograderSafeEnvError
from .Visibility import Visibility

def Test(*args, **kwargs):
    """
    The Test decorator is meant to signify an autograder test case. It will be auto added to the list of test cases so the main autograder can run it.
    You will supply the arguments to the decorator just as if you were calling the AutograderTest function WITHOUT the function. For example, if I wanted to 
    make a function with a name of "foo" with max score of 2, I could have:

    @Test("Foo", 2):
    def mytest(ag, test: AutograderTest):
        return 2

    Note: your function still MUST accept two positional arguments, the first gives you the main Autograder class while the second gives you the current
    tests AutograderTest class. That will be the method for interacting with the autograder.
    """
    def inner(func):
        AutograderTest(func, *args, **kwargs)
        return func
    return inner

def Setup(*args, **kwargs):
    """
    This is a decorator which will allow you to add a global setup function. 
    """
    def inner(func):
        AutograderSetup(func, *args, **kwargs)
        return func
    return inner

def Teardown(*args, **kwargs):
    """
    This is a decorator which will allow you have a global teardown function.
    """
    def inner(func):
        AutograderTeardown(func, *args, **kwargs)
        return func
    return inner

def SubTest(test: AutograderTest, *args, **kwargs):
    """
    This decorator will create a subtest out of a function.
    """
    def inner(func):
        AutograderSubTest(test, func, *args, **kwargs)
        return func
    return inner

__all__ = [
    "Autograder",
    "RateLimit",
    "AutograderTest",
    "Visibility",
    "Max",
    "global_tests",
    "Test",
    "Setup",
    "Teardown",
    "AutograderSubTest",
    "SubTest",
    "SubTestRunner",
    "StopSubTestRunner",
    "AutograderBaseError", 
    "AutograderFormatError", 
    "AutograderSafeEnvError"
]


"""Running tests"""
from __future__ import print_function

import sys
import time
import json

from unittest import result
from unittest.signals import registerResult


class JSONTestResult(result.TestResult):
    """A test result class that can print formatted text results to a stream.

    Used by JSONTestRunner.
    """
    def __init__(self, stream, descriptions, verbosity, results, leaderboard):
        super(JSONTestResult, self).__init__(stream, descriptions, verbosity)
        self.descriptions = descriptions
        self.results = results
        self.leaderboard = leaderboard

    def getDescription(self, test):
        doc_first_line = test.shortDescription()
        if self.descriptions and doc_first_line:
            return doc_first_line
        else:
            return str(test)

    def getTags(self, test):
        return getattr(getattr(test, test._testMethodName), '__tags__', None)

    def getWeight(self, test):
        return getattr(getattr(test, test._testMethodName), '__weight__', 0.0)

    def getScore(self, test):
        return getattr(getattr(test, test._testMethodName), '__score__', None)

    def getNumber(self, test):
        return getattr(getattr(test, test._testMethodName), '__number__', None)

    def getVisibility(self, test):
        return getattr(getattr(test, test._testMethodName), '__visibility__', None)

    def getLeaderboardData(self, test):
        column_name = getattr(getattr(test, test._testMethodName), '__leaderboard_column__', None)
        sort_order = getattr(getattr(test, test._testMethodName), '__leaderboard_sort_order__', None)
        value = getattr(getattr(test, test._testMethodName), '__leaderboard_value__', None)
        return (column_name, sort_order, value)

    def startTest(self, test):
        super(JSONTestResult, self).startTest(test)

    def getOutput(self):
        if self.buffer:
            out = self._stdout_buffer.getvalue()
            err = self._stderr_buffer.getvalue()
            if err:
                if not out.endswith('\n'):
                    out += '\n'
                out += err
            return out

    def buildResult(self, test, err=None):
        passed = (err == None)

        weight = self.getWeight(test)
        tags = self.getTags(test)
        number = self.getNumber(test)
        visibility = self.getVisibility(test)
        score = self.getScore(test)
        if score is None:
            score = weight if passed else 0.0

        output = self.getOutput()
        if err:
            output += "Test Failed: {0}\n".format(err[1])
        result = {
            "name": self.getDescription(test),
            "score": score,
            "max_score": weight,
        }
        if tags:
            result["tags"] = tags
        if output and len(output) > 0:
            result["output"] = output
        if visibility:
            result["visibility"] = visibility
        if number:
            result["number"] = number
        return result

    def buildLeaderboardEntry(self, test):
        name, sort_order, value = self.getLeaderboardData(test)
        return {
            "name": name,
            "value": value,
            "order": sort_order,
        }

    def processResult(self, test, err=None):
        if self.getLeaderboardData(test)[0]:
            self.leaderboard.append(self.buildLeaderboardEntry(test))
        else:
            self.results.append(self.buildResult(test, err))

    def addSuccess(self, test):
        super(JSONTestResult, self).addSuccess(test)
        self.processResult(test)

    def addError(self, test, err):
        super(JSONTestResult, self).addError(test, err)
        # Prevent output from being printed to stdout on failure
        self._mirrorOutput = False
        self.processResult(test, err)

    def addFailure(self, test, err):
        super(JSONTestResult, self).addFailure(test, err)
        self._mirrorOutput = False
        self.processResult(test, err)


class JSONTestRunner(object):
    """A test runner class that displays results in JSON form.
    """
    resultclass = JSONTestResult

    def __init__(self, stream=sys.stdout, descriptions=True, verbosity=1,
                 failfast=False, buffer=True, visibility=None,
                 stdout_visibility=None, post_processor=None):
        """
        Set buffer to True to include test output in JSON


        post_processor: if supplied, will be called with the final JSON
        data before it is written, allowing the caller to overwrite the
        test results (e.g. add a late penalty) by editing the results
        dict in the first argument.
        """
        self.stream = stream
        self.descriptions = descriptions
        self.verbosity = verbosity
        self.failfast = failfast
        self.buffer = buffer
        self.post_processor = post_processor
        self.json_data = {}
        self.json_data["tests"] = []
        self.json_data["leaderboard"] = []
        if visibility:
            self.json_data["visibility"] = visibility
        if stdout_visibility:
            self.json_data["stdout_visibility"] = stdout_visibility

    def _makeResult(self):
        return self.resultclass(self.stream, self.descriptions, self.verbosity,
                                self.json_data["tests"], self.json_data["leaderboard"])

    def run(self, test):
        "Run the given test case or test suite."
        result = self._makeResult()
        registerResult(result)
        result.failfast = self.failfast
        result.buffer = self.buffer
        startTime = time.time()
        startTestRun = getattr(result, 'startTestRun', None)
        if startTestRun is not None:
            startTestRun()
        try:
            test(result)
        finally:
            stopTestRun = getattr(result, 'stopTestRun', None)
            if stopTestRun is not None:
                stopTestRun()
        stopTime = time.time()
        timeTaken = stopTime - startTime

        self.json_data["execution_time"] = format(timeTaken, "0.2f")

        total_score = 0
        for test in self.json_data["tests"]:
            total_score += test["score"]
        self.json_data["score"] = total_score

        if self.post_processor is not None:
            self.post_processor(self.json_data)

        json.dump(self.json_data, self.stream, indent=4)
        self.stream.write('\n')
        return result



"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:03:47 
 * @Last Modified by:   ThaumicMekanism [Stephan K.] 
 * @Last Modified time: 2020-01-23 21:03:47 
 */
"""
import signal

class Timeout():
    """Timeout class using ALARM signal."""
    class Timeout(Exception):
        pass
 
    def __init__(self, sec):
        self.sec = sec
 
    def __enter__(self):
        if self.sec is not None:
            signal.signal(signal.SIGALRM, self.raise_timeout)
            signal.alarm(self.sec)
 
    def __exit__(self, *args):
        if self.sec is not None:
            signal.alarm(0)    # disable alarm
 
    def raise_timeout(self, *args):
        raise Timeout.Timeout()


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 20:57:36 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-04-25 14:56:44
 */
"""
"""
This is the base of the autograder.
"""
import datetime
import importlib
import json
import os
from typing import Callable, List, Union

from .AutograderErrors import AutograderSafeEnvError, AutograderHalt
from .AutograderLeaderboard import Leaderboard
from .AutograderRateLimit import RateLimit
from .AutograderSetup import global_setups
from .AutograderTeardown import global_teardowns
from .AutograderTest import AutograderTest, global_tests, Max
from .Utils import root_dir, submission_dir, results_path, get_welcome_message, is_local, WhenToRun, submission_metadata_dir

printed_welcome_message = False

class Autograder:
    use_ratelimit_when_local = False

    def __init__(self, *,
        rate_limit: RateLimit=None, 
        reverse_tests: bool=False, 
        export_tests_after_test: bool=True, 
        modify_results=lambda results: results,
        print_welcome_message: bool=True,
    ):
        if print_welcome_message:
            global printed_welcome_message
            if not printed_welcome_message:
                printed_welcome_message = True
                print(get_welcome_message())
        self.tests = []
        self.setups = []
        self.teardowns = []
        self.results_file = results_path()
        self.score = None
        self.output = None
        self.visibility = None
        self.stdout_visibility = None
        self.extra_data = {}
        self.leaderboard = Leaderboard()
        self.reverse_tests = reverse_tests
        self.export_tests_after_test = export_tests_after_test
        # rate_limit takes in a RateLimit class.
        # reset_time is when you want to reset the submission time. You
        # can leave it out to ignore. Put the time stirng in this format:
        #  "2018-11-29T16:15:00"
        if rate_limit is None:
            rate_limit = RateLimit()
        self.rate_limit: RateLimit = rate_limit
        self.start_time = datetime.datetime.now()
        self.modify_results = modify_results

        if not is_local():
            with open(submission_metadata_dir(), "r") as jsonMetadata:
                self.metadata = json.load(jsonMetadata)
            self.extra_data["id"] = self.metadata["id"]
        else:
            if os.path.isfile(submission_metadata_dir()):
                with open(submission_metadata_dir(), "r") as jsonMetadata:
                    self.metadata = json.load(jsonMetadata)
                self.extra_data["id"] = self.metadata["id"]
            else:
                self.extra_data["id"] = "LOCAL"
                self.metadata = None

    def run(self, import_globals: bool=True):
        def load_and_execute_autograder(ag: "Autograder"):
            if import_globals:
                for t in global_tests:
                    ag.add_test(t)
                for s in global_setups:
                    ag.add_setup(s)
                for t in global_teardowns:
                    ag.add_teardown(t)
            ag.execute()
        return self.safe_main(load_and_execute_autograder)

    def safe_main(self, f: Callable[["Autograder"], None]):
        def handler(exception):
            self.ag_fail("An exception occured in the autograder's main function. Please contact a TA to resolve this issue.")
            return True
        def wrapper():
            return f(self)
        return self.safe_env(wrapper, handler)

    def dump_results(self, data: dict) -> None:
        jsondata = json.dumps(data, ensure_ascii=False)
        with open(self.results_file, "wb") as f:
            f.write(jsondata.encode("ascii", errors="backslashreplace"))
        return self

    def import_tests(self, *, 
        tests_dir: Union[str, List[str]]=None, 
        test_files: Union[str, List[str]]=None, 
        blacklist: Union[str, List[str]]=None,
        verbose: bool=True
    ):
        if tests_dir is None:
            tests_dir = []
        if isinstance(tests_dir, str):
            tests_dir = [tests_dir]

        if test_files is None:
            test_files = []
        if isinstance(test_files, str):
            test_files = [test_files]
        
        def import_file(filename, package=None):
            if verbose:
                print(f"Importing file {filename} in package {package}.")
            file = filename
            if file.endswith(".py"):
                file = file[:-3]   
            try:
                importlib.import_module(".{}".format(file), package=package)
            except Exception as e:
                print(f"Could not add a test file {filename}!")
                import traceback
                traceback.print_exc()
                print(e)
        
        for test_file in test_files:
            import_file(test_file)
        
        for dir in tests_dir:
            files = sorted(os.listdir(dir), reverse=True)
            for file in files:
                filepath = os.path.join(dir, file)
                if file.endswith(".py") and file not in blacklist and os.path.isfile(filepath):
                    import_file(file, package=dir)
        return self

    def add_test(self, test, index=None):
        if isinstance(test, AutograderTest):
            if index is None:
                self.tests.append(test)
            else:
                self.tests.insert(index, test)
        else:
            raise ValueError("You must add type Test to the autograder.")
        return self

    def add_setup(self, setupfn):
        self.setups.append(setupfn)
        return self

    def add_teardown(self, teardownfn):
        self.teardowns.append(teardownfn)
        return self

    def set_score(self, score):
        self.score = score
        return self
    
    def add_score(self, addition):
        if self.score is None:
            self.score = 0
        self.score += addition
        return self
    
    def get_score(self):
        score = None
        for test in self.tests:
            test_score = test.get_score()
            if test_score is not None:
                if score is None:
                    score = test_score
                else:
                    score += test_score
        if score is None:
            score = self.score
        return score

    def print(self, *args, sep=' ', end='\n', file=None, flush=True, also_stdout=False):
        if self.output is None:
            self.output = ""
        msg = sep.join(map(str, args)) + end
        if also_stdout:
            print(msg)
        self.output += msg
        return self

    def create_test(self, *args, **kwargs):
        test = AutograderTest(*args, **kwargs)
        self.add_test(test)
        return self

    def ag_fail(self, message: str, extra: dict={}, exit_prog=True) -> None:
        data = {
            "score": 0,
            "output": message
        }
        data.update(extra)
        self.dump_results(data)
        if exit_prog:
            import sys
            sys.exit()
        return self
    
    def safe_env(self, f, handler=None):
        try:
            return f()
        except Exception as exc:
            print("An exception occured in the safe environment!")
            import traceback
            traceback.print_exc()
            print(exc)
            if handler is not None:
                try:
                    h = handler(exc)
                    if h:
                        return AutograderSafeEnvError(h)
                except Exception as exc:
                    print("An exception occurred while executing the exception handler!")
                    traceback.print_exc()
            self.ag_fail("An unexpected exception ocurred while trying to execute the autograder. Please try again or contact a TA if this persists.")
            return AutograderSafeEnvError(exc)

    def run_tests(self):
        global printed_welcome_message
        if not printed_welcome_message:
            printed_welcome_message = True
            print(get_welcome_message())
        local = is_local()
        def handle_failed():
                self.set_score(0)
                if "sub_counts" in self.extra_data:
                    self.print("[Rate Limit]: Since the autograder failed to run, you will not use up a token!")
                    self.rate_limit.rate_limit_unset_submission(self)
        for setup in self.setups:
            if not setup.when_to_run.okay_to_run(local):
                continue
            res = setup.run(self)
            if not res:
                print(f"[Error]: ({setup.name}) Returned non-true value `{res}` so assuming it failed!")
                self.print("[Error]: An error occurred in the setup of the Autograder!")
                handle_failed()
                return False
        for test in self.tests:
            test.run(self)
            if self.export_tests_after_test:
                self.generate_results(print_main_score_warning_error=False)
        for teardown in self.teardowns:
            if not teardown.when_to_run.okay_to_run(local):
                continue
            res = teardown.run(self)
            if not res:
                print(f"[Error]: ({teardown.name}) Returned non-true value `{res}` so assuming it failed!")
                self.print("[Error]: An error occurred in the teardown of the Autograder!")
                handle_failed()
                return False
        return True

    def generate_results(self, test_results=None, leaderboard: Leaderboard=None, dump=True, print_main_score_warning_error=True):
        results = {
            "execution_time": (datetime.datetime.now() - self.start_time).total_seconds(),
        }
        if test_results is None:
            tests = []
            if self.reverse_tests:
                tsts = reversed(self.tests)
            else:
                tsts = self.tests
            for test in tsts:
                res = test.get_results()
                if res:
                    tests.append(res)
            if tests:
                results["tests"] = tests
        else:
            if isinstance(test_results, list):
                results["tests"] = test_results
        if self.score is not None:
            results["score"] = self.score
        else:
            if "tests" not in results or len(results["tests"]) == 0 or not any(["score" in t for t in results["tests"]]):
                results["score"] = 0
                if print_main_score_warning_error:
                    self.print("This autograder does not set the main score or have any tests which give points!")
        if self.output is not None:
            results["output"] = self.output
        if self.visibility is not None:
            results["visibility"] = self.visibility
        if self.stdout_visibility is not None:
            results["stdout_visibility"] = self.stdout_visibility
        if self.extra_data:
            results["extra_data"] = self.extra_data
        if leaderboard is not None:
            leaderboard_export = leaderboard.export()
        else:
            leaderboard_export = self.leaderboard.export()
        if leaderboard_export:
            results["leaderboard"] = leaderboard_export
        results = self.modify_results(results)
        if dump:
            self.dump_results(results)
        return results
        
    def execute(self, generate_results: bool=True):
        global printed_welcome_message
        if not printed_welcome_message:
            printed_welcome_message = True
            print(get_welcome_message())
        self.rate_limit.rate_limit_main(self)
        if not self.run_tests():
            print("An error has occurred when attempting to run all tests.")
        if self.rate_limit.is_enabled():
            if self.output is None:
                self.output = ""
            self.output = self.rate_limit.get_rate_limit_str(self) + self.output
        if generate_results:
            self.generate_results()
        return self

    @staticmethod
    def root_dir() -> str:
        return root_dir()

    @staticmethod
    def submission_dir() -> str:
        return submission_dir()

    @staticmethod
    def DUMP(msg):
        ag = Autograder()
        ag.print(msg)
        ag.set_score(0)
        ag.generate_results()


"""
/*
 * @Author: ThaumicMekanism [Stephan K.]
 * @Date: 2020-01-30 11:35:10
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-30 16:09:58
 */
"""
class AutograderBaseError(Exception):
    """
    Base exception for autograders.
    """
    def __init__(self, info: any=None):
        self.info = info

class AutograderFormatError(AutograderBaseError):
    """
    This error is for when the expected state of the autograder is incorrect.
    """
    pass

class AutograderSafeEnvError(AutograderBaseError):
    pass

class AutograderHalt(AutograderBaseError):
    pass


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:02:51 
 * @Last Modified by:   ThaumicMekanism [Stephan K.] 
 * @Last Modified time: 2020-01-23 21:02:51 
 */
"""
from .Timeout import Timeout
from .Utils import WhenToRun

global_teardowns = []

class AutograderTeardown:
    def __init__(self, teardownfn, name, timeout: int=None, when_to_run: WhenToRun=WhenToRun.BOTH):
        self.teardownfn = teardownfn
        self.name = name
        self.timeout = timeout
        self.when_to_run = when_to_run
        global_teardowns.append(self)

    def run(self, ag: "Autograder"):
        success = False
        def f():
            nonlocal success
            try:
                with Timeout(self.timeout):
                    success = self.teardownfn(ag)
            except Timeout.Timeout:
                print(f"The teardown {self.name} has timed out!")
                ag.print("[ERROR]: A teardown step timed out!")
                success = False

        def handler(exception):
            nonlocal success
            success = False
            print(f"The teardown {self.name} has encountered an error!")
            ag.print("[Error]: An unexpected error occured in the Autograder when attempting to run a teardown of the Autograder! Please contact a TA if this persists.")

        ag.safe_env(f, handler=handler)
        return success
    



"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:04:08 
 * @Last Modified by:   ThaumicMekanism [Stephan K.] 
 * @Last Modified time: 2020-01-23 21:04:08 
 */
"""
"""
This is the visibility of an assignment.
"""
from enum import Enum

class Visibility(Enum):
    hidden = "hidden"
    after_due_date = "after_due_date"
    after_published = "after_published"
    visible = "visible"



import os

SUBMISSION_BASE = '/autograder/submission'


def check_submitted_files(paths, base=SUBMISSION_BASE):
    """Checks that the files in the given list exist in the student's submission.

    Returns a list of missing files.

    eg. check_submitted_files(['src/calculator.py'])
    """
    missing_files = []
    for path in paths:
        target_path = os.path.join(base, path)
        if not os.path.isfile(target_path):
            missing_files.append(path)
    return missing_files



"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:03:21 
 * @Last Modified by:   ThaumicMekanism [Stephan K.] 
 * @Last Modified time: 2020-01-23 21:03:21 
 */
"""
from typing import TypeVar
T = TypeVar('T')

class Leaderboard:
    def __init__(self):
        self.items = {}

    def add_item(self, name: str, value: T, order: str=None):
        """
        The default order is descending ('desc') though it also accepts 'asc' for ascending order.
        """
        self.items[name] = LeaderboardItem(name, value, order)
    
    def get_item(self, name: str):
        return self.items
    
    def remove_item(self, name: str):
        if name in self.items:
            del self.items[name]
            return True
        return False

    def export(self):
        return [leaderboard_item.export() for leaderboard_item in self.items]


class LeaderboardItem:
    def __init__(self, name: str, value: T, order: str=None) -> None:
        self.name = name
        self.value = value
        self.order = order

    def export(self):
        item = {
            "name": self.name,
            "value": self.value
        }
        if self.order is not None:
            item["order"] = self.order
        return item


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:02:33 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-30 16:17:58
 */
"""
from .AutograderTest import AutograderTest
from .Autograder import Autograder, AutograderSafeEnvError
from . import Visibility
from typing import Callable
from .AutograderErrors import AutograderFormatError
from .Utils import NoneLooseVersion

SUB_TESTS_KEY = "sub_tests"
ISPASSFAIL = True

class AutograderSubTest(AutograderTest):
    def __init__(
        self,
        test: AutograderTest,
        test_fn: Callable[["Autograder", AutograderTest], any]=None,
        name: str=None, 
        max_score: float=None,
        score: float=None,
        number: str=None, 
        tags: [str]=None,
        visibility: Visibility=None,
        extra_data=None,
        kill_autograder_on_error: bool=False,
        do_not_set_score: bool=False,
        timeout: int=None,
        ceil: bool=True,
        floor: bool=True,
        do_not_override_test_fn: bool=False,
    ):
        self.test = test
        self.test_fn = test_fn
        self.max_score = max_score
        self.number = number
        self.name = name
        self.tags = tags
        self.extra_data = extra_data
        self.score = score
        self.output = ""
        self.kill_autograder_on_error = kill_autograder_on_error
        self.do_not_set_score = do_not_set_score
        self.timeout = timeout
        self.ceil = ceil
        self.floor = floor
        test_case_has_test_runner_fn = False
        if issubclass(type(test.test_fn), SubTestRunner):
            test_case_has_test_runner_fn = True
        if test.test_fn is not None and not test_case_has_test_runner_fn:
            raise ValueError("The test case given already has a custom subtest function! If you do not want to override it, please set do_not_override_test_fn to true.")
        if test.test_fn is None and do_not_override_test_fn:
            raise ValueError("The test function for the test case does not have a test function!")
        if not do_not_override_test_fn and not test_case_has_test_runner_fn:
            test.test_fn = SubTestRunner()
        if SUB_TESTS_KEY not in test.__dict__:
            test.__dict__[SUB_TESTS_KEY] = []
        test.__dict__[SUB_TESTS_KEY].append(self)

    def set_score(self, score):
        if type(score) is bool:
            self.score = score
            return
        return super().set_score(score)

    def passed(self):
        if self.score is True or self.score == self.max_score:
            return True
        return False

    def get_score(self, convert_true_false=True):
        score = self.score
        if convert_true_false:
            if score is True:
                score = self.max_score
            elif score is False:
                score = 0
        if score is None:
            return 0
        return score

class StopSubTestRunner(BaseException):
    def __init__(self, info=""):
        self.info = info
        
class SubTestRunner(object):
    def __init__(
        self,
        is_pass_fail: bool=True,
        pass_fail_ratio: float=1,
    ):
        self.separator_count = 40
        self.is_pass_fail = is_pass_fail
        self.pass_fail_ratio = pass_fail_ratio

    def pre_test_run(self, ag: Autograder, test: AutograderTest, data):
        pass

    def get_sub_tests(self, ag: Autograder, test: AutograderTest, data):
        if SUB_TESTS_KEY not in test.__dict__:
            raise AutograderFormatError("This test should have test cases but none exist!")
        sub_test = test.__dict__[SUB_TESTS_KEY]
        sub_test = sorted(sub_test, key=lambda st: NoneLooseVersion(st.number))
        return sub_test

    def score_post(self, ag: AutograderTest, test: AutograderTest, data):
        test.print()
        test.print("+" * self.separator_count)
        if self.is_pass_fail:
            if self.did_pass(data):
                test.set_score(True)
                passed = "[✓] PASSED"
            else:
                test.set_score(False)
                passed = "[X] FAILED"
            test.print(passed)
        else:
            test.set_score(data["score"])
            test.print(f"{data['score']} / {test.max_score}")
        test.print("+" * self.separator_count)

    def post_test_run(self, ag: Autograder, test: AutograderTest, data):
        pass

    def pre_subtest_run(self, ag: Autograder, test: AutograderTest, t: AutograderSubTest, data):
        test.print("=" * self.separator_count)
        test.print(f"[SubTest]: {t.name}")
        test.print("-" * self.separator_count)

    def post_subtest_run(self, ag: Autograder, test: AutograderTest, t: AutograderSubTest, data):
        passed = t.passed()
        score = t.get_score()
        self.increment_score(data, score)
        self.add_test_passed_status(data, passed)
        test.print(t.output)
        test.print("-" * self.separator_count)
        if self.is_pass_fail:
            msg = "[X] FAILED"
            if passed:
                msg = "[✓] PASSED"
            test.print(msg)
        else:
            test.print(f"{score} / {t.max_score}")
        test.print("_" * self.separator_count)
        test.print("\n")

    def increment_score(self, data, score):
        data["score"] += score

    def get_full_score(self, data, score):
        return data["score"]

    def add_test_passed_status(self, data, status):
        data["passed"].append(status)

    def did_pass(self, data):
        amt = len(data["passed"])
        if amt == 0:
            return False
        return sum(data["passed"]) / amt >= self.pass_fail_ratio

    def run_test(self, ag: Autograder, test: AutograderTest, t, data):
        try:
            res = t.run(ag, handler=self.stopSubTestRunnerHandler)
        except StopSubTestRunner as e:
            if e.info:
                test.print(e.info)
            return e
        if isinstance(res, AutograderSafeEnvError):
            return res.info

    def part_stopped(self, idx: str, ag: Autograder, test: AutograderTest, t: AutograderSubTest, data: dict):
        test.print("[Warning]: This test stopped early!")
        return StopSubTestRunner(idx)

    def __call__(self, ag: Autograder, test: AutograderTest):
        r = self.runner(ag, test)
        if isinstance(r, StopSubTestRunner):
            print(f"[Warning]: The subtest runner stopped on {r.info}! ({test.name})")

    @staticmethod
    def stopSubTestRunnerHandler(exception):
        if isinstance(exception, StopSubTestRunner):
            return exception

    def runner(self, ag: Autograder, test: AutograderTest):
        data = {
            "score": 0,
            "passed": [],
        }
        sub_test = self.get_sub_tests(ag, test, data)
        if sub_test is False:
            r = self.part_stopped("get_sub_tests", ag, test, None, data)
            if isinstance(r, StopSubTestRunner):
                return r
        if self.pre_test_run(ag, test, data) is False:
            r = self.part_stopped("pre_test_run", ag, test, None, data)
            if isinstance(r, StopSubTestRunner):
                return r
        for t in sub_test:
            if self.pre_subtest_run(ag, test, t, data) is False:
                r = self.part_stopped("pre_subtest_run", ag, test, t, data)
                if isinstance(r, StopSubTestRunner):
                    return r
            res = self.run_test(ag, test, t, data)
            is_subtest_stopper = isinstance(res, StopSubTestRunner)
            if res is False or is_subtest_stopper:
                r = self.part_stopped("run_test", ag, test, t, data)
                if isinstance(r, StopSubTestRunner):
                    return r
            if self.post_subtest_run(ag, test, t, data) is False:
                r = self.part_stopped("post_subtest_run", ag, test, t, data)
                if isinstance(r, StopSubTestRunner):
                    return r
        if self.score_post(ag, test, data) is False:
            r = self.part_stopped("score_post", ag, test, None, data)
            if isinstance(r, StopSubTestRunner):
                return r
        if self.post_test_run(ag, test, data) is False:
            r = self.part_stopped("post_test_run", ag, test, None, data)
            if isinstance(r, StopSubTestRunner):
                return r



"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:03:57 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-30 16:27:10
 */
"""
import enum
import os
import pathlib

from distutils.version import LooseVersion

with open(os.path.join(pathlib.Path(__file__).parent.absolute(), 'VERSION')) as version_file:
    VERSION = version_file.read().strip()

def get_welcome_message():
    return f"Initializing the GradescopeBase Autograder v{VERSION} created by ThaumicMekanism [Stephan K.]."

def is_local() -> bool:
    return os.environ.get("IS_LOCAL") == "true"

def root_dir() -> str:
    """
    This function assumes the root directory is the one right above the GradescopeBase folder.
    """
    dirname = os.path.dirname
    return dirname(dirname(os.path.realpath(__file__)))

def submission_dir() -> str:
    """
    This returns the dir which contains the submission.
    """
    if is_local():
        return "./submission"
    return "/autograder/submission"

def submission_metadata_dir() -> str:
    """
    This returns the dir which contains the submission.
    """
    if is_local():
        return "./submission_metadata.json"
    return "/autograder/submission_metadata.json"

def results_path() -> str:
    """
    This returns the path which the results json should be exported to.
    """
    if is_local():
        return "./results/results.json"
    return "/autograder/results/results.json"

class NoneLooseVersion(LooseVersion):
    def __init__ (self, vstring=None):
        if vstring:
            self.parse(vstring)
        else:
            self.version = None

    def _cmp (self, other):
        if isinstance(other, str):
            other = NoneLooseVersion(other)
        if self.version == other.version:
            return 0
        if self.version is not None:
            return -1
        if other.version is not None:
            return 1
        if self.version < other.version:
            return -1
        if self.version > other.version:
            return 1

def merge(a, b, path=None):
    "merges b into a"
    if path is None: path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass # same leaf value
            else:
                raise Exception('Conflict at %s' % '.'.join(path + [str(key)]))
        else:
            a[key] = b[key]
    return a

class WhenToRun(enum.Enum):
    NEITHER = 0
    LOCAL = 1
    GRADESCOPE = 2
    BOTH = 3
    
    def okay_to_run(self, state: bool=None) -> bool:
        if state is None:
            state = is_local()
        if self is self.NEITHER:
            return False
        if self is self.BOTH:
            return True
        return (state and self is self.LOCAL) or (not state and self is self.GRADESCOPE)


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:01:33 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-23 21:02:27
 */
 """
from . import Autograder, AutograderTest, AutograderSubTest

class AutograderState:
    def __init__(self, ag: Autograder, test: Autograder, subtest: AutograderSubTest=None):
        self.ag = ag
        self.test = test
        self.subtest = subtest


"""
/*
 * @Author: ThaumicMekanism [Stephan K.] 
 * @Date: 2020-01-23 21:03:09 
 * @Last Modified by: ThaumicMekanism [Stephan K.]
 * @Last Modified time: 2020-01-30 16:04:03
 */
"""

"""
This is a test in gradescope.
"""
from .Timeout import Timeout
from . import Visibility
from .Utils import root_dir, submission_dir
from typing import Callable

global_tests = []

class Max:
    pass

default_visibility = None
class AutograderTest:
    def __init__(self, 
        test_fn: Callable[["Autograder", "AutograderTest"], any]=None,
        name: str=None, 
        max_score: float=None, 
        score: float=None,
        number: str=None, 
        tags: [str]=None, 
        visibility: Visibility=default_visibility, 
        extra_data=None,
        kill_autograder_on_error: bool=False,
        do_not_set_score: bool=False,
        timeout: int=None,
        ceil: bool=True,
        floor: bool=True,
        ran: bool=False,
    ):
        """
        The test_fn MUST take in parameters Autograder and AutograderTest in that order.
        It is how the test will interact with the autograder.
        """
        self.test_fn = test_fn
        self.max_score = max_score
        self.name = name
        self.number = number
        self.tags = tags
        self.visibility = visibility
        self.extra_data = extra_data
        self.score = score
        self.output = ""
        self.kill_autograder_on_error = kill_autograder_on_error
        self.do_not_set_score = do_not_set_score
        self.timeout = timeout
        self.ceil = ceil
        self.floor = floor
        self.ran = ran
        global_tests.append(self)

    def print(self, *args, sep=' ', end='\n', file=None, flush=True, also_stdout=False):
        msg = sep.join(map(str, args)) + end
        if also_stdout:
            print(msg)
        self.output += msg

    def set_score(self, score):
        if score is None:
            return
        if score is True:
            score = self.max_score
        elif score is False:
            score = 0
        if isinstance(score, Max) and self.max_score is not None:
            score = self.max_score
        # if self.score is None:
        #     self.score = score
        #     return
        if self.ceil and self.max_score is not None and score > self.max_score:
            score = self.max_score
        if self.floor and score < 0:
            score = 0
        self.score = score
    
    def remove_score(self):
        self.score = None

    def get_score(self):
        return self.score

    def run(self, ag, handler=None):
        self.ran = True
        if self.test_fn is None:
            self.print("[ERROR]: This test case does not have a callable function!")
            self.set_score(0)
            return
        
        def set_score(r):
            if self.do_not_set_score:
                return
            self.set_score(r)

        def f():
            try:
                with Timeout(self.timeout):
                    r = self.test_fn(ag, self)
                    set_score(r)
            except Timeout.Timeout:
                self.print("[ERROR]: This test timed out!")
                set_score(0)

        def default_handler(exception):
            if not self.do_not_set_score:
                self.set_score(0)
            if self.kill_autograder_on_error:
                return False
            self.print("[Error]: An unexpected error occured in the Autograder when attempting to run this testcase! Please contact a TA if this persists.")
            return True
        ag.safe_env(f, handler=handler if handler is not None else default_handler)

    def get_results(self):
        o = self.output
        if self.ran is False:
            o = "[WARNING]: This test did not run!"
        data = {"output": o}
        if self.max_score is not None:
            data["max_score"] = self.max_score
        if self.name is not None:
            data["name"] = self.name
        if self.number is not None:
            data["number"] = self.number
        if isinstance(self.tags, list):
            data["tags"] = self.tags
        if self.visibility is not None:
            data["visibility"] = self.visibility.name
        if self.extra_data is not None:
            data["extra_data"] = self.extra_data
        if self.score is not None:
            data["score"] = self.score
        return data

    @staticmethod
    def root_dir() -> str:
        return root_dir()

    @staticmethod
    def submission_dir() -> str:
        return submission_dir()